/*
 * Copyright (c) 2010 Tomasz Krasuski (see AUTHORS)
 *
 * This file is part of the A8CAS project which allows to manipulate tape
 * images for Atari 8-bit computers.
 *
 * A8CAS is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * A8CAS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with A8CAS; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 */
#include "denoise.h"

void DENOISE_init(DENOISE_t *denoise, double min_length)
{
	denoise->min_length = min_length;
	denoise->prev_length = 0.0;
	denoise->prev_signal = 0;
}

int DENOISE_next(DENOISE_t *denoise, char signal, double *length)
{
	if (signal == denoise->prev_signal) {
		if (denoise->prev_length == 0.0) {
			/* Previous signal's length already returned */
			return 1;
		} else {
			double new_length = denoise->prev_length + *length;
			if (new_length >= denoise->min_length) {
				denoise->prev_length = 0.0;
				*length = new_length;
				return 1;
			} else {
				denoise->prev_length = new_length;
				return 0;
			}
		}
	} else {
		denoise->prev_signal = signal;
		if (denoise->prev_length == 0.0) {
			/* Previous signal's length already returned */
			if (*length >= denoise->min_length) {
				return 1;
			} else {
				denoise->prev_length = *length;
				return 0;
			}
		} else {
			double new_length = denoise->prev_length + *length;
			if (new_length >= denoise->min_length) {
				denoise->prev_length = 0.0;
				*length = new_length;
				return 1;
			} else {
				denoise->prev_length = new_length;
				return 0;
			}
		}
	}
}
