/*
 * Copyright (c) 2010 Tomasz Krasuski (see AUTHORS)
 *
 * This file is part of the A8CAS project which allows to manipulate tape
 * images for Atari 8-bit computers.
 *
 * A8CAS is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * A8CAS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with A8CAS; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 */
#include <string.h>

#include "fsk_demod.h"
#include "downsample.h"

#define MIN_THRESHOLD 3

/*#define SND_DEBUG*/
#ifdef SND_DEBUG
#include <sndfile.h>
static SNDFILE *snd1, *snd2, *snd3;
static short int buf1[1024];
static float buf2[1024], buf3[1024];
static unsigned int pos = 0;

#endif

void FSK_DEMOD_init(FSK_DEMOD_t *demod, unsigned int samplerate)
{
	DOWNSAMPLE_init(&demod->downsample, samplerate, FSK_DEMOD_downsample_rate);

	demod->prev_signal = 1;
	demod->threshold = 0;

	demod->index = 0;
	memset(demod->history, 0, sizeof (demod->history));
	memset(demod->history_sum3, 0, sizeof (demod->history_sum3));
	memset(demod->history_sum4, 0, sizeof (demod->history_sum4));
#ifdef SND_DEBUG
	{
		SF_INFO info;
		info.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
		info.samplerate = FSK_DEMOD_downsample_rate;
		info.channels = 1;
		snd1 = sf_open("snd1.wav", SFM_WRITE, &info);
		snd2 = sf_open("snd2.wav", SFM_WRITE, &info);
		snd3 = sf_open("snd3.wav", SFM_WRITE, &info);
	}
#endif
}

static int advance_downsampled(FSK_DEMOD_t *demod, float sample)
{
	int signal;
	int base = (demod->index++) & 31;
	int write = (base + 31) & 31;

	demod->history[write] = demod->history[write + 32] = sample;

#if 0
	{
	float const *xs = &demod->history[base];

	float const sin0_8 = 0.0f;
	float const sin1_8 = 0.707107f;
	float const sin2_8 = 1.0f;
	float const sin3_8 = 0.707107f;
	float const sin4_8 = 0.0f;
	float const sin5_8 = -0.707107f;
	float const sin6_8 = -1.0f;
	float const sin7_8 = -0.707107f;

	float x0 = xs[0] + xs[ 8] + xs[16];
	float x1 = xs[1] + xs[ 9] + xs[17];
	float x2 = xs[2] + xs[10] + xs[18];
	float x3 = xs[3] + xs[11] + xs[19];
	float x4 = xs[4] + xs[12] + xs[20];
	float x5 = xs[5] + xs[13] + xs[21];
	float x6 = xs[6] + xs[14] + xs[22];
	float x7 = xs[7] + xs[15] + xs[23];

	float zero_s = x0*sin0_8 + x1*sin1_8 + x2*sin2_8 + x3*sin3_8 + x4*sin4_8 + x5*sin5_8 + x6*sin6_8 + x7*sin7_8;
	float zero_c = x0*sin2_8 + x1*sin3_8 + x2*sin4_8 + x3*sin5_8 + x4*sin6_8 + x5*sin7_8 + x6*sin0_8 + x7*sin1_8;

	float const sin0_6 = 0.0f;
	float const sin1_6 = 0.86602540378443864676372317075294f;
	float const sin2_6 = 0.86602540378443864676372317075294f;
	float const sin3_6 = 0.0f;
	float const sin4_6 = -0.86602540378443864676372317075294f;
	float const sin5_6 = -0.86602540378443864676372317075294f;
	float const cos0_6 = 1.0f;
	float const cos1_6 = 0.5f;
	float const cos2_6 = -0.5f;
	float const cos3_6 = -1.0f;
	float const cos4_6 = -0.5f;
	float const cos5_6 = 0.5f;

	float y0 = xs[0] + xs[ 6] + xs[12] + xs[18];
	float y1 = xs[1] + xs[ 7] + xs[13] + xs[19];
	float y2 = xs[2] + xs[ 8] + xs[14] + xs[20];
	float y3 = xs[3] + xs[ 9] + xs[15] + xs[21];
	float y4 = xs[4] + xs[10] + xs[16] + xs[22];
	float y5 = xs[5] + xs[11] + xs[17] + xs[23];

	float one_s = y0*sin0_6 + y1*sin1_6 + y2*sin2_6 + y3*sin3_6 + y4*sin4_6 + y5*sin5_6;
	float one_c = y0*cos0_6 + y1*cos1_6 + y2*cos2_6 + y3*cos3_6 + y4*cos4_6 + y5*cos5_6;
#else
	{
	float const *xs = &demod->history[base + 8];
	demod->history_sum3[write] = demod->history_sum3[write + 32] = xs[7] + xs[15] + xs[23];
	demod->history_sum4[write] = demod->history_sum4[write + 32] = xs[5] + xs[11] + xs[17] + xs[23];
	}

	{
	float const *s3 = &demod->history_sum3[base + 32 - 8];
	float const *s4 = &demod->history_sum4[base + 32 - 6];

	float const x0 = s3[0];
	float const x1 = s3[1];
	float const x2 = s3[2];
	float const x3 = s3[3];
	float const x4 = s3[4];
	float const x5 = s3[5];
	float const x6 = s3[6];
	float const x7 = s3[7];

	float const sqrt2 = 0.707107f;
	float x1mx5 = x1 - x5;
	float x3mx7 = x3 - x7;
	float zero_s = (x2 - x6) + (x1mx5 + x3mx7)*sqrt2;
	float zero_c = (x0 - x4) + (x1mx5 - x3mx7)*sqrt2;

	float const y1 = s4[0];
	float const y2 = s4[1];
	float const y3 = s4[2];
	float const y4 = s4[3];
	float const y5 = s4[4];
	float const y6 = s4[5];

	float const half = 0.5f;
	float const sin2_12 = 0.86602540378443864676372317075294f;

	float y2my5 = y2 - y5;
	float y3my6 = y3 - y6;
	float one_s = (y2my5 + y3my6)*sin2_12;
	float one_c = (y1 - y4) + (y2my5 - y3my6)*half;
#endif

	float one = (one_s * one_s + one_c * one_c) * ((8.0f / 6.0f) * (8.0f / 6.0f));
	float zero = (zero_s * zero_s + zero_c * zero_c);

	signal = (one > zero);
/*	signal = (sample > 0);*/
/*	{
		float xmin = xs[0], xmax = xs[0];
		unsigned int i;
		for (i = 1; i < 24; i ++) {
			if (xs[i] < xmin)
				xmin = xs[i];
			if (xs[i] > xmax)
				xmax = xs[i];
		}
		signal = zero < ((xmax - xmin) * 5);
	}*/
	if (signal != demod->prev_signal) {
		if (demod->threshold < MIN_THRESHOLD) {
			demod->threshold ++;
			signal = demod->prev_signal;
		} else {
			demod->threshold = 0;
			demod->prev_signal = signal;
		}
	} else {
		if (demod->threshold > 0) {
			demod->threshold --;
		}
	}


#ifdef SND_DEBUG
	{
		buf1[pos ++] = signal * 32000 - 16000;
		if (pos == 1024) {
			sf_write_short(snd1, buf1, 1024);
			sf_write_float(snd2, buf2, 1024);
			sf_write_float(snd3, buf3, 1024);
			pos = 0;
		}
	}
#endif
	return signal;
	}
}

int FSK_DEMOD_advance(FSK_DEMOD_t *demod, short int sample)
{
	if (DOWNSAMPLE_next(&demod->downsample, &sample) != 0)
		return advance_downsampled(demod, ((float)sample + 0.5)/ 32767.5);
	else
		return demod->prev_signal;
}
